@@ -5,12 +5,15 @@ encourage it by way of some defaults. Rather, they are ways of building
 arbitrary Grafana JSON.
 """
 
-import attr
-from attr.validators import instance_of, in_
 import itertools
 import math
-from numbers import Number
+
+import string
 import warnings
+from numbers import Number
+
+import attr
+from attr.validators import in_, instance_of
 
 
 @attr.s
@@ -71,6 +74,7 @@ FLOT = 'flot'
 ABSOLUTE_TYPE = 'absolute'
 DASHBOARD_TYPE = 'dashboard'
 GRAPH_TYPE = 'graph'
+STAT_TYPE = 'stat'
 SINGLESTAT_TYPE = 'singlestat'
 TABLE_TYPE = 'table'
 TEXT_TYPE = 'text'
@@ -79,6 +83,8 @@ BARGAUGE_TYPE = "bargauge"
 GAUGE_TYPE = "gauge"
 HEATMAP_TYPE = "heatmap"
 STATUSMAP_TYPE = "flant-statusmap-panel"
+SVG_TYPE = 'marcuscalidus-svg-panel'
+PIE_CHART_TYPE = 'grafana-piechart-panel'
 
 DEFAULT_FILL = 1
 DEFAULT_REFRESH = '10s'
@@ -111,6 +117,26 @@ SHORT_FORMAT = "short"
 BYTES_FORMAT = "bytes"
 BITS_PER_SEC_FORMAT = "bps"
 BYTES_PER_SEC_FORMAT = "Bps"
+NONE_FORMAT = "none"
+JOULE_FORMAT = "joule"
+WATTHOUR_FORMAT = "watth"
+WATT_FORMAT = "watt"
+KWATT_FORMAT = "kwatt"
+KWATTHOUR_FORMAT = "kwatth"
+VOLT_FORMAT = "volt"
+BAR_FORMAT = "pressurebar"
+PSI_FORMAT = "pressurepsi"
+CELSIUS_FORMAT = "celsius"
+KELVIN_FORMAT = "kelvin"
+GRAM_FORMAT = "massg"
+EUR_FORMAT = "currencyEUR"
+USD_FORMAT = "currencyUSD"
+METER_FORMAT = "lengthm"
+SQUARE_METER_FORMAT = "areaM2"
+CUBIC_METER_FORMAT = "m3"
+LITRE_FORMAT = "litre"
+PERCENT_FORMAT = "percent"
+VOLT_AMPERE_FORMAT = "voltamp"
 
 # Alert rule state
 STATE_NO_DATA = "no_data"
@@ -540,6 +566,20 @@ class Annotations(object):
         }
 
 
+@attr.s
+class DataLink(object):
+    title = attr.ib()
+    linkUrl = attr.ib(default="", validator=instance_of(str))
+    isNewTab = attr.ib(default=False, validator=instance_of(bool))
+
+    def to_json_data(self):
+        return {
+            'title': self.title,
+            'url': self.linkUrl,
+            'targetBlank': self.isNewTab,
+        }
+
+
 @attr.s
 class DataSourceInput(object):
     name = attr.ib()
@@ -1013,6 +1053,7 @@ class Graph(object):
     """
     Generates Graph panel json structure.
 
+    :param dataLinks: list of data links hooked to datapoints on the graph
     :param dataSource: DataSource's name
     :param minSpan: Minimum width for each panel
     :param repeat: Template's name to repeat Graph on
@@ -1022,6 +1063,7 @@ class Graph(object):
     targets = attr.ib()
     aliasColors = attr.ib(default=attr.Factory(dict))
     bars = attr.ib(default=False, validator=instance_of(bool))
+    dataLinks = attr.ib(default=attr.Factory(list))
     dataSource = attr.ib(default=None)
     description = attr.ib(default=None)
     editable = attr.ib(default=True, validator=instance_of(bool))
@@ -1082,6 +1124,9 @@ class Graph(object):
             'links': self.links,
             'minSpan': self.minSpan,
             'nullPointMode': self.nullPointMode,
+            'options': {
+                'dataLinks': self.dataLinks,
+            },
             'percentage': self.percentage,
             'pointradius': self.pointRadius,
             'points': self.points,
@@ -1105,6 +1150,36 @@ class Graph(object):
             graphObject['alert'] = self.alert
         return graphObject
 
+    def _iter_targets(self):
+        for target in self.targets:
+            yield target
+
+    def _map_targets(self, f):
+        return attr.assoc(self, targets=[f(t) for t in self.targets])
+
+    def auto_ref_ids(self):
+        """Give unique IDs all the panels without IDs.
+
+        Returns a new ``Graph`` that is the same as this one, except all of
+        the metrics have their ``refId`` property set. Any panels which had
+        an ``refId`` property set will keep that property, all others will
+        have auto-generated IDs provided for them.
+        """
+        ref_ids = set([t.refId for t in self._iter_targets() if t.refId])
+        double_candidate_refs = \
+            [p[0] + p[1] for p
+                in itertools.product(string.ascii_uppercase, repeat=2)]
+        candidate_ref_ids = itertools.chain(
+            string.ascii_uppercase,
+            double_candidate_refs,
+        )
+
+        auto_ref_ids = (i for i in candidate_ref_ids if i not in ref_ids)
+
+        def set_refid(t):
+            return t if t.refId else attr.assoc(t, refId=next(auto_ref_ids))
+        return self._map_targets(set_refid)
+
 
 @attr.s
 class SparkLine(object):
@@ -1239,11 +1314,179 @@ class AlertList(object):
         }
 
 
+@attr.s
+class Stat(object):
+    """Generates Stat panel json structure
+
+    Grafana doc on stat: https://grafana.com/docs/grafana/latest/panels/visualizations/stat-panel/
+
+    :param dataSource: Grafana datasource name
+    :param targets: list of metric requests for chosen datasource
+    :param title: panel title
+    :param colorMode: defines if Grafana will color panel background: keys "value" "background"
+    :param graphMode: defines if Grafana will draw graph: keys 'area' 'none'
+    :param orientation: Stacking direction in case of multiple series or fields: keys 'auto' 'horizontal' 'vertical'
+    :param alignment: defines value & title positioning: keys 'auto' 'centre'
+    :param description: optional panel description
+    :param editable: defines if panel is editable via web interfaces
+    :param format: defines value units
+    :param height: defines panel height
+    :param id: panel id
+    :param decimals: number of decimals to display
+    :param interval: defines time interval between metric queries
+    :param links: additional web links
+    :param mappings: the list of values to text mappings
+        This should be a list of StatMapping objects
+        https://grafana.com/docs/grafana/latest/panels/field-configuration-options/#value-mapping
+    :param reduceCalc: algorithm for reduction to a single value: keys
+        'mean' 'lastNotNull' 'last' 'first' 'firstNotNull' 'min' 'max' 'sum' 'total'
+    :param span: defines the number of spans that will be used for panel
+    :param thresholds: single stat thresholds
+    :param transparent: defines if the panel should be transparent
+    """
+
+    dataSource = attr.ib()
+    targets = attr.ib()
+    title = attr.ib()
+    description = attr.ib(default=None)
+    colorMode = attr.ib(default='value')
+    graphMode = attr.ib(default='area')
+    orientation = attr.ib(default='auto')
+    alignment = attr.ib(default='auto')
+    editable = attr.ib(default=True, validator=instance_of(bool))
+    format = attr.ib(default='none')
+    height = attr.ib(default=None)
+    id = attr.ib(default=None)
+    links = attr.ib(default=attr.Factory(list))
+    mappings = attr.ib(default=attr.Factory(list))
+    span = attr.ib(default=6)
+    thresholds = attr.ib(default='')
+    timeFrom = attr.ib(default=None)
+    transparent = attr.ib(default=False, validator=instance_of(bool))
+    reduceCalc = attr.ib(default='mean', type=str)
+    decimals = attr.ib(default=None)
+
+    def to_json_data(self):
+        return {
+            'datasource': self.dataSource,
+            'description': self.description,
+            'editable': self.editable,
+            'id': self.id,
+            'links': self.links,
+            'height': self.height,
+            'fieldConfig': {
+                'defaults': {
+                    'custom': {},
+                    'decimals': self.decimals,
+                    'mappings': self.mappings,
+                    'thresholds': {
+                        'mode': 'absolute',
+                        'steps': self.thresholds,
+                    },
+                    'unit': self.format
+                }
+            },
+            'options': {
+                'colorMode': self.colorMode,
+                'graphMode': self.graphMode,
+                'justifyMode': self.alignment,
+                'orientation': self.orientation,
+                'reduceOptions': {
+                    'calcs': [
+                        self.reduceCalc
+                    ],
+                    'fields': '',
+                    'values': False
+                }
+            },
+            'span': self.span,
+            'targets': self.targets,
+            'title': self.title,
+            'transparent': self.transparent,
+            'type': STAT_TYPE,
+            'timeFrom': self.timeFrom,
+        }
+
+
+@attr.s
+class StatMapping(object):
+    """
+    Generates json structure for the value mapping for the Stat panel:
+    :param text: Sting that will replace input value
+    :param value: Value to be replaced
+    :param startValue: When using a range, the start value of the range
+    :param endValue: When using a range, the end value of the range
+    :param id: panel id
+    """
+
+    text = attr.ib()
+    mapValue = attr.ib(default="", validator=instance_of(str))
+    startValue = attr.ib(default="", validator=instance_of(str))
+    endValue = attr.ib(default="", validator=instance_of(str))
+    id = attr.ib(default=None)
+
+    def to_json_data(self):
+        mappingType = MAPPING_TYPE_VALUE_TO_TEXT if self.mapValue else MAPPING_TYPE_RANGE_TO_TEXT
+
+        return {
+            'operator': "",
+            'text': self.text,
+            'type': mappingType,
+            'value': self.mapValue,
+            'from': self.startValue,
+            'to': self.endValue,
+            'id': self.id
+        }
+
+
+@attr.s
+class StatValueMapping(object):
+    """
+    Generates json structure for the value mappings for the StatPanel:
+    :param text: Sting that will replace input value
+    :param value: Value to be replaced
+    :param id: panel id
+    """
+
+    text = attr.ib()
+    mapValue = attr.ib(default="", validator=instance_of(str))
+    id = attr.ib(default=None)
+
+    def to_json_data(self):
+        return StatMapping(self.text, mapValue=self.mapValue, id=self.id)
+
+
+@attr.s
+class StatRangeMapping(object):
+    """
+    Generates json structure for the value mappings for the StatPanel:
+    :param text: Sting that will replace input value
+    :param startValue: When using a range, the start value of the range
+    :param endValue: When using a range, the end value of the range
+    :param id: panel id
+    """
+
+    text = attr.ib()
+    startValue = attr.ib(default="", validator=instance_of(str))
+    endValue = attr.ib(default="", validator=instance_of(str))
+    id = attr.ib(default=None)
+
+    def to_json_data(self):
+        return StatMapping(
+            self.text,
+            startValue=self.startValue,
+            endValue=self.endValue,
+            id=self.id
+        )
+
+
 @attr.s
 class SingleStat(object):
     """Generates Single Stat panel json structure
 
-    Grafana doc on singlestat: http://docs.grafana.org/reference/singlestat/
+    This panel was deprecated in Grafana 7.0, please use Stat instead
+
+    Grafana doc on singlestat: https://grafana.com/docs/grafana/latest/features/panels/singlestat/
 
     :param dataSource: Grafana datasource name
     :param targets: list of metric requests for chosen datasource
@@ -1251,7 +1494,7 @@ class SingleStat(object):
     :param cacheTimeout: metric query result cache ttl
     :param colors: the list of colors that can be used for coloring
         panel value or background. Additional info on coloring in docs:
-        http://docs.grafana.org/reference/singlestat/#coloring
+        https://grafana.com/docs/grafana/latest/features/panels/singlestat/#coloring
     :param colorBackground: defines if grafana will color panel background
     :param colorValue: defines if grafana will color panel value
     :param description: optional panel description
@@ -1266,7 +1509,7 @@ class SingleStat(object):
     :param links: additional web links
     :param mappingType: defines panel mapping type.
         Additional info can be found in docs:
-        http://docs.grafana.org/reference/singlestat/#value-to-text-mapping
+        https://grafana.com/docs/grafana/latest/features/panels/singlestat/#value-to-text-mapping
     :param mappingTypes: the list of available mapping types for panel
     :param maxDataPoints: maximum metric query results,
         that will be used for rendering
@@ -1281,7 +1524,7 @@ class SingleStat(object):
     :param span: defines the number of spans that will be used for panel
     :param sparkline: defines if grafana should draw an additional sparkline.
         Sparkline grafana documentation:
-        http://docs.grafana.org/reference/singlestat/#spark-lines
+        https://grafana.com/docs/grafana/latest/features/panels/singlestat/#spark-lines
     :param thresholds: single stat thresholds
     :param transparent: defines if panel should be transparent
     :param valueFontSize: defines value font size
@@ -1663,27 +1906,6 @@ class Table(object):
         }
 
 
-@attr.s
-class Threshold(object):
-    """Threshold for a gauge
-
-    :param color: color of threshold
-    :param index: index of color in gauge
-    :param value: when to use this color will be null if index is 0
-    """
-
-    color = attr.ib()
-    index = attr.ib(validator=instance_of(int))
-    value = attr.ib(validator=instance_of(float))
-
-    def to_json_data(self):
-        return {
-            "color": self.color,
-            "index": self.index,
-            "value": "null" if self.index == 0 else self.value,
-        }
-
-
 @attr.s
 class BarGauge(object):
     """Generates Bar Gauge panel json structure
@@ -1691,6 +1913,7 @@ class BarGauge(object):
     :param allValue: If All values should be shown or a Calculation
     :param cacheTimeout: metric query result cache ttl
     :param calc: Calculation to perform on metrics
+    :param dataLinks: list of data links hooked to datapoints on the graph
     :param dataSource: Grafana datasource name
     :param decimals: override automatic decimal precision for legend/tooltips
     :param description: optional panel description
@@ -1728,6 +1951,7 @@ class BarGauge(object):
     allValues = attr.ib(default=False, validator=instance_of(bool))
     cacheTimeout = attr.ib(default=None)
     calc = attr.ib(default=GAUGE_CALC_MEAN)
+    dataLinks = attr.ib(default=attr.Factory(list))
     dataSource = attr.ib(default=None)
     decimals = attr.ib(default=None)
     description = attr.ib(default=None)
@@ -1800,6 +2024,7 @@ class BarGauge(object):
                         "min": self.min,
                         "title": self.label,
                         "unit": self.format,
+                        "links": self.dataLinks,
                     },
                     "limit": self.limit,
                     "mappings": self.valueMaps,
@@ -1829,6 +2054,7 @@ class GaugePanel(object):
     :param allValue: If All values should be shown or a Calculation
     :param cacheTimeout: metric query result cache ttl
     :param calc: Calculation to perform on metrics
+    :param dataLinks: list of data links hooked to datapoints on the graph
     :param dataSource: Grafana datasource name
     :param decimals: override automatic decimal precision for legend/tooltips
     :param description: optional panel description
@@ -1843,7 +2069,7 @@ class GaugePanel(object):
     :param links: additional web links
     :param max: maximum value of the gauge
     :param maxDataPoints: maximum metric query results,
-           that will be used for rendering
+        that will be used for rendering
     :param min: minimum value of the gauge
     :param minSpan: minimum span number
     :param rangeMaps: the list of value to text mappings
@@ -1864,6 +2090,7 @@ class GaugePanel(object):
     allValues = attr.ib(default=False, validator=instance_of(bool))
     cacheTimeout = attr.ib(default=None)
     calc = attr.ib(default=GAUGE_CALC_MEAN)
+    dataLinks = attr.ib(default=attr.Factory(list))
     dataSource = attr.ib(default=None)
     decimals = attr.ib(default=None)
     description = attr.ib(default=None)
@@ -1921,6 +2148,7 @@ class GaugePanel(object):
                         "min": self.min,
                         "title": self.label,
                         "unit": self.format,
+                        "links": self.dataLinks,
                     },
                     "limit": self.limit,
                     "mappings": self.valueMaps,
@@ -1992,6 +2220,7 @@ class Heatmap(object):
     :param yBucketNumber
     :param highlightCards: boolean
     :param hideZeroBuckets: boolean
+    :param transparent: defines if the panel should be transparent
     """
 
     title = attr.ib()
@@ -2025,6 +2254,7 @@ class Heatmap(object):
     hideZeroBuckets = attr.ib(default=False)
     highlightCards = attr.ib(default=True)
     options = attr.ib(default=None)
+    transparent = attr.ib(default=False, validator=instance_of(bool))
 
     xAxis = attr.ib(
         default=attr.Factory(XAxis),
@@ -2061,6 +2291,7 @@ class Heatmap(object):
             'targets': self.targets,
             'title': self.title,
             'tooltip': self.tooltip,
+            'transparent': self.transparent,
             'type': HEATMAP_TYPE,
             'xAxis': self.xAxis,
             'xBucketNumber': self.xBucketNumber,
@@ -2208,3 +2439,185 @@ class Statusmap(object):
         if self.alert:
             graphObject['alert'] = self.alert
         return graphObject
+
+
+@attr.s
+class Svg(object):
+    """Generates SVG panel json structure
+    Grafana doc on SVG: https://grafana.com/grafana/plugins/marcuscalidus-svg-panel
+    :param dataSource: Grafana datasource name
+    :param targets: list of metric requests for chosen datasource
+    :param title: panel title
+    :param description: optional panel description
+    :param editable: defines if panel is editable via web interfaces
+    :param format: defines value units
+    :param jsCodeFilePath: path to javascript file to be run on dashboard refresh
+    :param jsCodeInitFilePath: path to javascript file to be run after the first initialization of the SVG
+    :param height: defines panel height
+    :param id: panel id
+    :param interval: defines time interval between metric queries
+    :param links: additional web links
+    :param reduceCalc: algorithm for reduction to a single value: keys 'mean' 'lastNotNull' 'last' 'first' 'firstNotNull' 'min' 'max' 'sum' 'total'
+    :param span: defines the number of spans that will be used for panel
+    :param svgFilePath: path to SVG image file to be displayed
+    """
+
+    dataSource = attr.ib()
+    targets = attr.ib()
+    title = attr.ib()
+    description = attr.ib(default=None)
+    editable = attr.ib(default=True, validator=instance_of(bool))
+    format = attr.ib(default="none")
+    jsCodeFilePath = attr.ib(default="", validator=instance_of(str))
+    jsCodeInitFilePath = attr.ib(default="", validator=instance_of(str))
+    height = attr.ib(default=None)
+    id = attr.ib(default=None)
+    links = attr.ib(default=attr.Factory(list))
+    span = attr.ib(default=6)
+    svgFilePath = attr.ib(default="", validator=instance_of(str))
+
+    @staticmethod
+    def read_file(file_path):
+        if file_path:
+            with open(file_path) as f:
+                read_data = f.read()
+            return read_data
+        else:
+            return ""
+
+    def to_json_data(self):
+
+        js_code = self.read_file(self.jsCodeFilePath)
+        js_init_code = self.read_file(self.jsCodeInitFilePath)
+        svg_data = self.read_file(self.svgFilePath)
+
+        return {
+            'datasource': self.dataSource,
+            'description': self.description,
+            'editable': self.editable,
+            'id': self.id,
+            'links': self.links,
+            'height': self.height,
+            'format': self.format,
+            'js_code': js_code,
+            'js_init_code': js_init_code,
+            'span': self.span,
+            'svg_data': svg_data,
+            'targets': self.targets,
+            'title': self.title,
+            'type': SVG_TYPE,
+            'useSVGBuilder': False
+        }
+
+
+@attr.s
+class PieChart(object):
+    """Generates Pie Chart panel json structure
+    Grafana doc on Pie Chart: https://grafana.com/grafana/plugins/grafana-piechart-panel
+    :param dataSource: Grafana datasource name
+    :param targets: list of metric requests for chosen datasource
+    :param title: panel title
+    :param description: optional panel description
+    :param editable: defines if panel is editable via web interfaces
+    :param format: defines value units
+    :param height: defines panel height
+    :param id: panel id
+    :param pieType: defines the shape of the pie chart (pie or donut)
+    :param showLegend: defines if the legend should be shown
+    :param showLegend: defines if the legend should show values
+    :param legendType: defines where the legend position
+    :param links: additional web links
+    :param span: defines the number of spans that will be used for panel
+    :param transparent: defines if the panel is transparent
+    """
+
+    dataSource = attr.ib()
+    targets = attr.ib()
+    title = attr.ib()
+    description = attr.ib(default=None)
+    editable = attr.ib(default=True, validator=instance_of(bool))
+    format = attr.ib(default='none')
+    height = attr.ib(default=None)
+    id = attr.ib(default=None)
+    links = attr.ib(default=attr.Factory(list))
+    legendType = attr.ib(default='Right side')
+    pieType = attr.ib(default='pie')
+    showLegend = attr.ib(default=True)
+    showLegendValues = attr.ib(default=True)
+    span = attr.ib(default=6)
+    thresholds = attr.ib(default='')
+    timeFrom = attr.ib(default=None)
+    transparent = attr.ib(default=False, validator=instance_of(bool))
+
+    def to_json_data(self):
+        return {
+            'datasource': self.dataSource,
+            'description': self.description,
+            'editable': self.editable,
+            'format': self.format,
+            'id': self.id,
+            'links': self.links,
+            'pieType': self.pieType,
+            'height': self.height,
+            'fieldConfig': {
+                'defaults': {
+                    'custom': {},
+                },
+                'overrides': []
+            },
+            'legend': {
+                'show': self.showLegend,
+                'values': self.showLegendValues
+            },
+            'legendType': self.legendType,
+            'span': self.span,
+            'targets': self.targets,
+            'title': self.title,
+            'type': PIE_CHART_TYPE,
+            'timeFrom': self.timeFrom,
+            'transparent': self.transparent
+        }
+
+
+@attr.s
+class Threshold(object):
+    """Threshold for a gauge
+
+    :param color: color of threshold
+    :param index: index of color in gauge
+    :param value: when to use this color will be null if index is 0
+    """
+
+    color = attr.ib()
+    index = attr.ib(validator=instance_of(int))
+    value = attr.ib(validator=instance_of(float))
+    line = attr.ib(default=True, validator=instance_of(bool))
+    op = attr.ib(default="gt")
+    yaxis = attr.ib(default="left")
+
+    def to_json_data(self):
+        return {
+            "op": self.op,
+            "yaxis": self.yaxis,
+            "color": self.color,
+            "line": self.line,
+            "index": self.index,
+            "value": "null" if self.index == 0 else self.value,
+        }
+
+
+class SeriesOverride(object):
+    alias = attr.ib()
+    bars = attr.ib(default=False)
+    lines = attr.ib(default=True)
+    yaxis = attr.ib(default=1)
+    color = attr.ib(default=None)
+
+    def to_json_data(self):
+        return {
+            "alias": self.alias,
+            "bars": self.bars,
+            "lines": self.lines,
+            "yaxis": self.yaxis,
+            "color": self.color,
+        }
